import numpy as np
from scipy.stats import entropy, itemfreq
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import LabelEncoder


def knn_purity(data, labels: np.ndarray, n_neighbors=30):
    """Computes KNN Purity for ``data`` given the labels.
        Parameters
        ----------
        data:
            Numpy ndarray of data
        labels
            Numpy ndarray of labels
        n_neighbors: int
            Number of nearest neighbors.
        Returns
        -------
        score: float
            KNN purity score. A float between 0 and 1.
    """
    labels = LabelEncoder().fit_transform(labels.ravel())

    nbrs = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(data)
    indices = nbrs.kneighbors(data, return_distance=False)[:, 1:]
    neighbors_labels = np.vectorize(lambda i: labels[i])(indices)

    # pre cell purity scores
    scores = ((neighbors_labels - labels.reshape(-1, 1)) == 0).mean(axis=1)
    res = [
        np.mean(scores[labels == i]) for i in np.unique(labels)
    ]  # per cell-type purity

    return np.mean(res)


def entropy_batch_mixing(data, labels,
                         n_neighbors=50, n_pools=50, n_samples_per_pool=100):
    """Computes Entory of Batch mixing metric for ``adata`` given the batch column name.
        Parameters
        ----------
        data
            Numpy ndarray of data
        labels
            Numpy ndarray of labels
        n_neighbors: int
            Number of nearest neighbors.
        n_pools: int
            Number of EBM computation which will be averaged.
        n_samples_per_pool: int
            Number of samples to be used in each pool of execution.
        Returns
        -------
        score: float
            EBM score. A float between zero and one.
    """

    def __entropy_from_indices(indices, n_cat):
        return entropy(np.array(itemfreq(indices)[:, 1].astype(np.int32)), base=n_cat)

    n_cat = len(np.unique(labels))
    # print(f'Calculating EBM with n_cat = {n_cat}')

    neighbors = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(data)
    indices = neighbors.kneighbors(data, return_distance=False)[:, 1:]
    batch_indices = np.vectorize(lambda i: labels[i])(indices)

    entropies = np.apply_along_axis(__entropy_from_indices, axis=1, arr=batch_indices, n_cat=n_cat)

    # average n_pools entropy results where each result is an average of n_samples_per_pool random samples.
    if n_pools == 1:
        score = np.mean(entropies)
    else:
        score = np.mean([
            np.mean(entropies[np.random.choice(len(entropies), size=n_samples_per_pool)])
            for _ in range(n_pools)
        ])

    return score



from sklearn.metrics import r2_score

import torch
import torch.distributions as db
import torch.nn as nn
from torch.distributions.kl import kl_divergence

from scvi.module.base import BaseModuleClass, auto_move_data
from scvi.nn import Encoder, FCLayers
from scvi.module import Classifier
from scvi import settings

from ._metrics import entropy_batch_mixing, knn_purity
from ._utils import DecoderNormal, DrugNetwork, VanillaEncoder, REGISTRY_KEYS

import numpy as np


class CPAModule(BaseModuleClass):
    """
    CPA module using Gaussian/NegativeBinomial Likelihood

    Parameters
    ----------
        n_genes: int
        n_treatments: int
        covars_encoder: dict
            Dictionary of covariates with keys as each covariate name and values as
                number of unique values of the corresponding covariate
        n_latent: int
            Latent Dimension
        loss_ae: str
            Autoencoder loss (either "gauss" or "mse")
        doser: str
            # TODO: What is this
        autoencoder_width: int
        autoencoder_depth: int
        use_batch_norm: bool
        use_layer_norm: bool
        variational: bool
    """

    def __init__(self,
                 n_genes: int,
                 n_drugs: int,
                 covars_encoder: dict,
                 n_latent: int = 128,
                 loss_ae="gauss",
                 doser_type="logsigm",
                 output_activation: str = 'linear',
                 autoencoder_width=256,
                 autoencoder_depth=3,
                 adversary_width=64,
                 adversary_depth=2,
                 dosers_width: int = 128,
                 dosers_depth: int = 2,
                 use_batch_norm: bool = True,
                 use_layer_norm: bool = False,
                 dropout_rate: float = 0.0,
                 variational: bool = False,
                 seed: int = 0,
                 ):
        super().__init__()

        assert loss_ae in ['mse', 'gauss']

        torch.manual_seed(seed)
        np.random.seed(seed)
        settings.seed = seed

        self.n_genes = n_genes
        self.n_drugs = n_drugs
        self.n_latent = n_latent
        self.loss_ae = loss_ae
        self.doser_type = doser_type
        self.ae_width = autoencoder_width
        self.ae_depth = autoencoder_depth
        self.dosers_width = dosers_width
        self.dosers_depth = dosers_depth
        self.adversary_width = adversary_width
        self.adversary_depth = adversary_depth
        self.use_batch_norm = use_batch_norm
        self.use_layer_norm = use_layer_norm
        self.variational = variational

        self.covars_encoder = covars_encoder

        self.variational = variational
        if variational:
            self.encoder = Encoder(
                n_genes,
                n_latent,
                var_activation=nn.Softplus(),
                n_hidden=autoencoder_width,
                n_layers=autoencoder_depth,
                use_batch_norm=use_batch_norm,
                use_layer_norm=use_layer_norm,
                dropout_rate=dropout_rate,
                activation_fn=nn.ReLU,
            )
        else:
            self.encoder = VanillaEncoder(
                n_input=n_genes,
                n_output=n_latent,
                n_hidden=autoencoder_width,
                n_layers=autoencoder_depth,
                use_batch_norm=use_batch_norm,
                use_layer_norm=use_layer_norm,
                dropout_rate=dropout_rate,
                activation_fn=nn.ReLU,
            )

        # Decoder components
        if loss_ae in ["gauss", 'mse']:
            self.decoder = DecoderNormal(
                n_input=n_latent,
                n_output=n_genes,
                n_hidden=autoencoder_width,
                n_layers=autoencoder_depth,
                use_batch_norm=use_batch_norm,
                use_layer_norm=use_layer_norm,
                output_activation=output_activation,
                dropout_rate=dropout_rate,
            )

        else:
            raise Exception('Invalid Loss function for Autoencoder')

        # Embeddings
        # 1. Drug Network
        self.drug_network = DrugNetwork(n_drugs=self.n_drugs,
                                        n_latent=self.n_latent,
                                        doser_type=self.doser_type,
                                        n_hidden=self.dosers_width,
                                        n_layers=self.dosers_depth,
                                        dropout_rate=dropout_rate,
                                        )

        self.drugs_classifier = Classifier(
            n_input=n_latent,
            n_labels=n_drugs,
            n_hidden=self.adversary_width,
            n_layers=self.adversary_depth,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            dropout_rate=dropout_rate,
            activation_fn=nn.ReLU,
            logits=True,
        )

        # 2. Covariates Embedding
        self.covars_embedding = nn.ModuleDict(
            {
                key: torch.nn.Embedding(len(unique_covars), n_latent)
                for key, unique_covars in self.covars_encoder.items()
            }
        )

        self.covars_classifiers = nn.ModuleDict(
            {
                key: Classifier(n_input=n_latent,
                                n_labels=len(unique_covars),
                                n_hidden=self.adversary_width,
                                n_layers=self.adversary_depth,
                                use_batch_norm=use_batch_norm,
                                use_layer_norm=use_layer_norm,
                                dropout_rate=dropout_rate,
                                logits=True)
                if len(unique_covars) > 1 else None

                for key, unique_covars in self.covars_encoder.items()
            }
        )

        self.ae_loss_fn = nn.GaussianNLLLoss()
        self.adv_loss_covariates = nn.CrossEntropyLoss()

        self.adv_loss_drugs = nn.BCEWithLogitsLoss()

    def _get_inference_input(self, tensors):
        x = tensors[REGISTRY_KEYS.X_KEY]  # batch_size, n_genes
        drugs_doses = tensors['drugs_doses']

        covars_dict = dict()
        for covar, unique_covars in self.covars_encoder.items():
            encoded_covars = tensors[covar].view(-1, )  # (batch_size,)
            covars_dict[covar] = encoded_covars

        input_dict = dict(
            genes=x,
            drugs=drugs_doses,
            doses=None,
            covars_dict=covars_dict,
        )
        return input_dict

    @auto_move_data
    def inference(
            self,
            genes,
            drugs,
            doses,
            covars_dict,
    ):
        # x_ = torch.log1p(x)
        batch_size = genes.shape[0]
        x_ = genes
        if self.variational:
            z_means, z_vars, latent_basal = self.encoder(x_)
            basal_distribution = db.Normal(z_means, z_vars.sqrt())
        else:
            basal_distribution = None
            latent_basal = self.encoder(x_)

        latent_treatment = self.drug_network(drugs, doses)

        latent_covariates = []
        for covar, _ in self.covars_encoder.items():
            latent_covar_i = self.covars_embedding[covar](covars_dict[covar].long())
            latent_covar_i = latent_covar_i.view(batch_size, self.n_latent).unsqueeze(0)  # 1, batch_size, n_latent
            latent_covariates.append(latent_covar_i)

        if len(latent_covariates) > 0:
            latent_covariates = torch.cat(latent_covariates, 0).sum(0)  # Summing all covariates representations
            latent = latent_basal + latent_treatment + latent_covariates
        else:
            latent = latent_basal + latent_treatment

        return dict(
            latent=latent,
            latent_basal=latent_basal,
            basal_distribution=basal_distribution,
        )

    def _get_generative_input(self, tensors, inference_outputs, **kwargs):
        input_dict = {}

        latent = inference_outputs["latent"]
        latent_basal = inference_outputs['latent_basal']

        covars_dict = dict()
        for covar, _ in self.covars_encoder.items():
            val = tensors[covar].view(-1, )
            covars_dict[covar] = val

        input_dict['latent'] = latent
        input_dict['latent_basal'] = latent_basal
        return input_dict

    @auto_move_data
    def generative(
            self,
            latent,
            latent_basal,
    ):
        outputs = self.decoder(inputs=latent)
        return dict(
            means=outputs.loc,
            variances=outputs.variance,
        )

    def loss(self, tensors, inference_outputs, generative_outputs):
        """Computes the reconstruction loss (AE) or the ELBO (VAE)"""
        x = tensors[REGISTRY_KEYS.X_KEY]

        means = generative_outputs["means"]
        variances = generative_outputs["variances"]

        if self.loss_ae in ["gauss", "mse"]:
            reconstruction_loss = self.ae_loss_fn(x, means, variances)
        else:
            raise Exception('Invalid Loss function for CPA')

        # TODO: Add KL annealing if needed
        kl_loss = 0.0
        if self.variational:
            basal_distribution = inference_outputs["basal_distribution"]
            dist_pz = db.Normal(
                torch.zeros_like(basal_distribution.loc), torch.ones_like(basal_distribution.scale)
            )
            kl_loss = kl_divergence(basal_distribution, dist_pz).sum(-1)

        return reconstruction_loss, kl_loss

    def adversarial_loss(self, tensors, latent_basal):
        """Computes adversarial classification losses and regularizations"""
        drugs_doses = tensors['drugs_doses']
        batch_size = drugs_doses.shape[0]

        covars_dict = dict()
        for covar, unique_covars in self.covars_encoder.items():
            encoded_covars = tensors[covar].view(-1, )  # (batch_size,)
            covars_dict[covar] = encoded_covars

        drugs_pred = self.drugs_classifier(latent_basal)
        covars_pred = {}
        for covar in self.covars_encoder.keys():
            if self.covars_classifiers[covar] is not None:
                covar_pred = self.covars_classifiers[covar](latent_basal)
                covars_pred[covar] = covar_pred
            else:
                covars_pred[covar] = None

        adv_results = {}
        # Classification losses for different covariates
        for covar in self.covars_encoder.keys():
            adv_results[f'adv_{covar}'] = self.adv_loss_covariates(
                covars_pred[covar],
                covars_dict[covar].long(),
            ) if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)
            adv_results[f'acc_{covar}'] = torch.sum(
                covars_pred[covar].argmax(1) == covars_dict[covar].long().squeeze(-1)) / batch_size \
                if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)

        # Classification loss for different drug combinations
        adv_results['adv_drugs'] = self.adv_loss_drugs(drugs_pred, drugs_doses.gt(
            0).float())
        adv_results['adv_loss'] = adv_results['adv_drugs'] + sum(
            [adv_results[f'adv_{key}'] for key in self.covars_encoder.keys()])

        # Penalty losses
        for covar in self.covars_encoder.keys():
            adv_results[f'penalty_{covar}'] = (
                torch.autograd.grad(
                    covars_pred[covar].sum(),
                    latent_basal,
                    create_graph=True
                )[0].pow(2).mean()
            ) if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)

        adv_results['penalty_drugs'] = (
            torch.autograd.grad(
                drugs_pred.sum(),
                latent_basal,
                create_graph=True,
            )[0].pow(2).mean()
        )
        adv_results['penalty_adv'] = adv_results['penalty_drugs'] + sum(
            [adv_results[f'penalty_{covar}'] for covar in self.covars_encoder.keys()])

        return adv_results

    def r2_metric(self, tensors, inference_outputs, generative_outputs):
        pred_mean = torch.nan_to_num(generative_outputs['means'], nan=1e2, neginf=-1e3,
                                     posinf=1e3).detach().cpu().numpy()  # batch_size, n_genes
        pred_var = torch.nan_to_num(generative_outputs['variances'], nan=1e2, neginf=-1e3,
                                    posinf=1e3).detach().cpu().numpy()  # batch_size, n_genes

        x = tensors[REGISTRY_KEYS.X_KEY].detach().cpu().numpy()  # batch_size, n_genes

        batch_size = x.shape[0]

        drugs_doses = tensors['drugs_doses']  # (batch_size, n_drugs)
        indices = (drugs_doses * torch.arange(drugs_doses.shape[1]).exp().view(1, -1).repeat(batch_size, 1).to(
            drugs_doses.device)).sum(dim=1)

        for i, covar in enumerate(self.covars_encoder.keys()):
            indices += tensors[covar].view(-1, )  # (batch_size,)

        r2_mean = 0.0
        r2_var = 0.0

        unique_indices = indices.unique()
        n_unique_indices = len(unique_indices)

        for index in unique_indices:
            index_mask = (indices == index).detach().cpu().numpy()
            x_index = x[index_mask]
            means_index = pred_mean[index_mask]
            variances_index = pred_var[index_mask]

            true_mean_index = x_index.mean(0)
            pred_mean_index = means_index.mean(0)

            true_var_index = x_index.var(0)
            pred_var_index = variances_index.mean(0)

            r2_mean += r2_score(true_mean_index, pred_mean_index) / n_unique_indices
            r2_var += r2_score(true_var_index, pred_var_index) / n_unique_indices

        return r2_mean, r2_var

    def disentanglement(self, tensors, inference_outputs, generative_outputs, linear=True):
        latent_basal = inference_outputs['latent_basal'].detach().cpu().numpy()
        latent = inference_outputs['latent'].detach().cpu().numpy()
        drugs_doses = tensors['drugs_doses']
        drug_names = drugs_doses.argmax(dim=1).float().detach().cpu().numpy()

        knn_basal = knn_purity(latent_basal, drug_names.ravel(), n_neighbors=min(drug_names.shape[0] - 1, 30))

        for covar, unique_covars in self.covars_encoder.items():
            if len(unique_covars) > 1:
                target_covars = tensors[f'{covar}'].detach().cpu().numpy()
                knn_basal += knn_purity(latent_basal, target_covars.ravel(),
                                        n_neighbors=min(target_covars.shape[0] - 1, 30))

        knn_after = knn_purity(latent, drug_names.ravel(), n_neighbors=min(drug_names.shape[0] - 1, 30))

        for covar, unique_covars in self.covars_encoder.items():
            if len(unique_covars) > 1:
                target_covars = tensors[f'{covar}'].detach().cpu().numpy()
                knn_after += knn_purity(latent, target_covars.ravel(), n_neighbors=min(target_covars.shape[0] - 1, 30))

        return knn_basal, knn_after

    def get_expression(self, tensors, **inference_kwargs):
        """Computes gene expression means and std.

        Only implemented for the gaussian likelihood.

        Parameters
        ----------
        tensors : dict
            Considered inputs

        """
        _, generative_outputs = self.forward(
            tensors,
            compute_loss=False,
            inference_kwargs=inference_kwargs,
        )
        if self.loss_ae in ["gauss", 'mse']:
            mus = torch.nan_to_num(generative_outputs['means'], nan=1e2, neginf=-1e3,
                                   posinf=1e3)  # batch_size, n_genes
            stds = torch.nan_to_num(generative_outputs['variances'], nan=1e2, neginf=-1e3,
                                    posinf=1e3)  # batch_size, n_genes
            return mus, stds
        else:
            raise ValueError

    def get_drug_embeddings(self, tensors, **inference_kwargs):
        inputs = self._get_inference_input(tensors)
        drugs = inputs['drugs']
        doses = None

        return self.drug_network(drugs, doses)



from typing import Union

import torch
from torch.optim.lr_scheduler import StepLR

from scvi.module.base import BaseModuleClass
from scvi.train import TrainingPlan

import numpy as np


class CPATrainingPlan(TrainingPlan):
    def __init__(
            self,
            module: BaseModuleClass,
            covars_to_ncovars: dict,
            autoencoder_lr=3e-4,
            n_steps_kl_warmup: Union[int, None] = None,
            n_epochs_kl_warmup: Union[int, None] = None,
            n_epochs_warmup: Union[int, None] = None,
            adversary_steps: int = 3,
            reg_adversary: float = 60,
            penalty_adversary: float = 60,
            dosers_lr=4e-3,
            dosers_wd=1e-7,
            adversary_lr=3e-4,
            adversary_wd=4e-7,
            autoencoder_wd=4e-7,
            step_size_lr: int = 45,
            batch_size: int = 256,
            variational: bool = False,
    ):
        """Training plan for the CPA model"""
        super().__init__(
            module=module,
            lr=autoencoder_lr,
            weight_decay=autoencoder_wd,
            n_steps_kl_warmup=n_steps_kl_warmup,
            n_epochs_kl_warmup=n_epochs_kl_warmup,
            reduce_lr_on_plateau=False,
            lr_factor=None,
            lr_patience=None,
            lr_threshold=None,
            lr_scheduler_metric=None,
            lr_min=None,
        )

        self.n_epochs_warmup = n_epochs_warmup if n_epochs_warmup is not None else 0

        self.covars_encoder = covars_to_ncovars

        self.autoencoder_wd = autoencoder_wd
        self.autoencoder_lr = autoencoder_lr

        self.adversary_lr = adversary_lr
        self.adversary_wd = adversary_wd
        self.adversary_steps = adversary_steps
        self.reg_adversary = reg_adversary
        self.penalty_adversary = penalty_adversary

        self.dosers_lr = dosers_lr
        self.dosers_wd = dosers_wd

        self.step_size_lr = step_size_lr

        self.batch_size = batch_size

        self.variational = variational

        self.automatic_optimization = False
        self.iter_count = 0

        self.epoch_history = {
            'mode': [],
            'epoch': [],
            'recon_loss': [],
            'adv_loss': [],
            'penalty_adv': [],
            'adv_drugs': [],
            'penalty_drugs': [],
            'reg_mean': [],
            'reg_var': [],
            'disent_basal': [],
            'disent_after': [],
        }

        for covar in self.covars_encoder.keys():
            self.epoch_history[f'adv_{covar}'] = []
            self.epoch_history[f'penalty_{covar}'] = []

    def configure_optimizers(self):
        if hasattr(self.module.drug_network, 'drug_encoder'):
            optimizer_autoencoder = torch.optim.Adam(
                list(filter(lambda p: p.requires_grad, self.module.encoder.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.decoder.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.drug_network.drug_embedding.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.drug_network.drug_encoder.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.covars_embedding.parameters())),
                lr=self.autoencoder_lr,
                weight_decay=self.autoencoder_wd)
        else:
            optimizer_autoencoder = torch.optim.Adam(
                list(filter(lambda p: p.requires_grad, self.module.encoder.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.decoder.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.drug_network.drug_embedding.parameters())) +
                list(filter(lambda p: p.requires_grad, self.module.covars_embedding.parameters())),
                lr=self.autoencoder_lr,
                weight_decay=self.autoencoder_wd)

        optimizer_adversaries = torch.optim.Adam(
            list(filter(lambda p: p.requires_grad, self.module.drugs_classifier.parameters())) +
            list(filter(lambda p: p.requires_grad, self.module.covars_classifiers.parameters())),
            lr=self.adversary_lr,
            weight_decay=self.adversary_wd)

        optimizer_dosers = torch.optim.Adam(
            filter(lambda p: p.requires_grad, self.module.drug_network.dosers.parameters()),
            lr=self.dosers_lr,
            weight_decay=self.dosers_wd)

        optimizers = [optimizer_autoencoder, optimizer_adversaries, optimizer_dosers]
        if self.step_size_lr is not None:
            scheduler_autoencoder = StepLR(optimizer_autoencoder, step_size=self.step_size_lr)
            scheduler_adversaries = StepLR(optimizer_adversaries, step_size=self.step_size_lr)
            scheduler_dosers = StepLR(optimizer_dosers, step_size=self.step_size_lr)
            schedulers = [scheduler_autoencoder, scheduler_adversaries, scheduler_dosers]
            return optimizers, schedulers
        else:
            return optimizers

    def training_step(self, batch, batch_idx):
        opt, opt_adv, opt_dosers = self.optimizers()

        inf_outputs, gen_outputs = self.module.forward(batch, compute_loss=False)

        if self.current_epoch >= self.n_epochs_warmup:
            # Adversarial update
            if self.iter_count % self.adversary_steps != 0:
                opt_adv.zero_grad()
                latent_basal = inf_outputs['latent_basal']
                adv_results = self.module.adversarial_loss(tensors=batch, latent_basal=latent_basal)
                self.manual_backward(adv_results['adv_loss'] + self.penalty_adversary * adv_results['penalty_adv'])
                opt_adv.step()

                for key, val in adv_results.items():
                    adv_results[key] = val.item()

                results = adv_results.copy()
                results.update({'recon_loss': 0.0})

            # Model update
            else:
                opt.zero_grad()
                opt_dosers.zero_grad()
                reconstruction_loss, kl_loss = self.module.loss(
                    tensors=batch,
                    inference_outputs=inf_outputs,
                    generative_outputs=gen_outputs,
                )
                latent_basal = inf_outputs['latent_basal']
                adv_results = self.module.adversarial_loss(tensors=batch, latent_basal=latent_basal)

                loss = reconstruction_loss - self.reg_adversary * adv_results['adv_loss']
                if self.variational:
                    loss += self.kl_weight * kl_loss.mean()
                self.manual_backward(loss)
                opt.step()
                opt_dosers.step()

                for key, val in adv_results.items():
                    adv_results[key] = val.item()

                results = adv_results.copy()

                results.update({'recon_loss': reconstruction_loss.item()})

        else:
            adv_results = {'adv_loss': 0.0, 'adv_drugs': 0.0, 'penalty_adv': 0.0, 'penalty_drugs': 0.0}
            for covar in self.covars_encoder.keys():
                adv_results[f'adv_{covar}'] = 0.0
                adv_results[f'penalty_{covar}'] = 0.0

            results = adv_results.copy()

            opt.zero_grad()
            opt_dosers.zero_grad()
            reconstruction_loss, kl_loss = self.module.loss(
                tensors=batch,
                inference_outputs=inf_outputs,
                generative_outputs=gen_outputs,
            )
            loss = reconstruction_loss
            if self.variational:
                loss += self.kl_weight * kl_loss.mean()
            self.manual_backward(loss)
            opt.step()
            opt_dosers.step()

            results.update({'recon_loss': reconstruction_loss.item()})

        self.iter_count += 1

        results.update({'reg_mean': 0.0, 'reg_var': 0.0})
        results.update({'disent_basal': 0.0})
        results.update({'disent_after': 0.0})

        return results

    def training_epoch_end(self, outputs):
        keys = ['recon_loss', 'adv_loss', 'penalty_adv', 'adv_drugs', 'penalty_drugs', 'reg_mean', 'reg_var',
                'disent_basal', 'disent_after']
        for key in keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))

        for covar in self.covars_encoder.keys():
            key1, key2 = f'adv_{covar}', f'penalty_{covar}'
            self.epoch_history[key1].append(np.mean([output[key1] for output in outputs]))
            self.epoch_history[key2].append(np.mean([output[key2] for output in outputs]))

        self.epoch_history['epoch'].append(self.current_epoch)
        self.epoch_history['mode'].append('train')

        self.log("recon", self.epoch_history['recon_loss'][-1], prog_bar=True)
        # self.log("adv_loss", self.epoch_history['adv_loss'][-1], prog_bar=True)
        # self.log("penalty_adv", self.epoch_history['penalty_adv'][-1], prog_bar=True)
        # self.log("reg_mean", self.epoch_history['reg_mean'][-1], prog_bar=True)
        # self.log("reg_var", self.epoch_history['reg_var'][-1], prog_bar=True)
        # self.log("disent_drugs", self.epoch_history['disent_drugs'][-1], prog_bar=True)

        if self.current_epoch > 1 and self.current_epoch % self.step_size_lr == 0:
            sch, sch_adv, sch_dosers = self.lr_schedulers()
            sch.step()
            sch_adv.step()
            sch_dosers.step()

    def get_progress_bar_dict(self):
        items = super().get_progress_bar_dict()
        items.pop('v_num')
        # items.pop('loss')
        return items

    def validation_step(self, batch, batch_idx):
        inf_outputs, gen_outputs = self.module.forward(batch, compute_loss=False)

        reconstruction_loss, kl_loss = self.module.loss(
            tensors=batch,
            inference_outputs=inf_outputs,
            generative_outputs=gen_outputs,
        )

        # if self.current_epoch >= self.n_epochs_warmup:
        #     adv_results = self.module.adversarial_loss(
        #         tensors=batch,
        #         inference_outputs=inf_outputs,
        #         generative_outputs=gen_outputs,
        #     )
        #     for key, val in adv_results.items():
        #         adv_results[key] = val.item()
        # else:
        adv_results = {'adv_loss': 0.0, 'adv_drugs': 0.0, 'penalty_adv': 0.0, 'penalty_drugs': 0.0}
        for covar in self.covars_encoder.keys():
            adv_results[f'adv_{covar}'] = 0.0
            adv_results[f'penalty_{covar}'] = 0.0

        r2_mean, r2_var = self.module.r2_metric(batch, inf_outputs, gen_outputs)
        disent_basal, disent_after = self.module.disentanglement(batch, inf_outputs, gen_outputs)

        results = adv_results
        results.update({'reg_mean': r2_mean, 'reg_var': r2_var})
        results.update({'disent_basal': disent_basal})
        results.update({'disent_after': disent_after})
        results.update({'recon_loss': reconstruction_loss.item()})
        results.update({'cpa_metric': r2_mean + r2_var + 1.0 + len(
            [covar for covar, unique_covars in self.covars_encoder.items() if
             len(unique_covars) > 1]) - disent_basal + disent_after})

        return results

    def validation_epoch_end(self, outputs):
        keys = ['recon_loss', 'adv_loss', 'penalty_adv', 'adv_drugs', 'penalty_drugs', 'reg_mean', 'reg_var',
                'disent_basal', 'disent_after']
        for key in keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))

        for covar in self.covars_encoder.keys():
            key1, key2 = f'adv_{covar}', f'penalty_{covar}'
            self.epoch_history[key1].append(np.mean([output[key1] for output in outputs]))
            self.epoch_history[key2].append(np.mean([output[key2] for output in outputs]))

        self.epoch_history['epoch'].append(self.current_epoch)
        self.epoch_history['mode'].append('valid')

        self.log('val_recon', self.epoch_history['recon_loss'][-1], prog_bar=True)
        self.log('cpa_metric', np.mean([output['cpa_metric'] for output in outputs]), prog_bar=False)
        self.log('val_disnt_basal', self.epoch_history['disent_basal'][-1], prog_bar=True)
        self.log('val_disnt_after', self.epoch_history['disent_after'][-1], prog_bar=True)
        self.log('val_reg_mean', self.epoch_history['reg_mean'][-1], prog_bar=True)
        # self.log('val_reg_var', self.epoch_history['reg_var'][-1], prog_bar=True)



import warnings

warnings.simplefilter('ignore')

from ._model import CPA
from ._module import CPAModule
from . import _plotting as pl
from ._api import ComPertAPI

try:
    import importlib.metadata as importlib_metadata
except:
    import importlib_metadata

package_name = "cpa-tools"
__version__ = importlib_metadata.version(package_name)

# __all__ = [
#     "CPA",
#     "CPAModule",
#     "ComPertAPI",
# ]



import torch
import torch.nn as nn
import torch.nn.functional as F

from scvi.nn import FCLayers
from torch.distributions import Normal


class _REGISTRY_KEYS:
    X_KEY: str = "X"
    PERTURBATIONS: str = "drugs_doses"
    CAT_COVS_KEY: str = "extra_categorical_covs"
    CONT_COVS_KEY: str = "extra_continuous_covs"
    INDICES_KEY: str = "ind_x"
    SIZE_FACTOR_KEY: str = "size_factor"
    DRUG_KEY: str = None
    DOSE_KEY: str = None
    CONTROL_KEY: str = None
    COVARS_KEYS: str = []


REGISTRY_KEYS = _REGISTRY_KEYS()


class VanillaEncoder(nn.Module):
    def __init__(
            self,
            n_input,
            n_output,
            n_hidden,
            n_layers,
            n_cat_list,
            use_layer_norm=True,
            use_batch_norm=False,
            output_activation: str = 'linear',
            dropout_rate: float = 0.1,
            activation_fn=nn.ReLU,
    ):
        super().__init__()
        self.n_output = n_output
        self.output_activation = output_activation

        self.network = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            use_layer_norm=use_layer_norm,
            use_batch_norm=use_batch_norm,
            dropout_rate=dropout_rate,
            activation_fn=activation_fn,
        )
        self.z = nn.Linear(n_hidden, n_output)

    def forward(self, inputs, *cat_list):
        z = self.z(self.network(inputs, *cat_list))
        return z


class DecoderNormal(nn.Module):
    def __init__(
            self,
            n_input,
            n_output,
            n_hidden,
            n_layers,
            n_cat_list,
            use_layer_norm=True,
            use_batch_norm=False,
            output_activation: str = 'linear',
            dropout_rate: float = 0.1,
    ):
        super().__init__()
        self.n_output = n_output
        self.output_activation = output_activation

        self.network = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            use_layer_norm=use_layer_norm,
            use_batch_norm=use_batch_norm,
            dropout_rate=dropout_rate,
            activation_fn=nn.ReLU,
        )
        self.mean = nn.Linear(n_hidden, n_output, bias=False)
        self.var = nn.Linear(n_hidden, n_output, bias=False)

    def forward(self, inputs, *cat_list):
        x = self.network(inputs, *cat_list)
        locs = self.mean(x)
        var_ = self.var(x)
        if self.output_activation == 'relu':
            locs = F.relu(locs)

        variances = var_.exp().add(1).log().add(1e-3)
        return Normal(loc=locs, scale=variances.sqrt())


class GeneralizedSigmoid(nn.Module):
    """
    Sigmoid, log-sigmoid or linear functions for encoding dose-response for
    drug perurbations.
    """

    def __init__(self, n_drugs, non_linearity='sigmoid'):
        """Sigmoid modeling of continuous variable.
        Params
        ------
        nonlin : str (default: logsigm)
            One of logsigm, sigm.
        """
        super(GeneralizedSigmoid, self).__init__()
        self.non_linearity = non_linearity
        self.n_drugs = n_drugs
        self.beta = torch.nn.Parameter(
            torch.ones(1, n_drugs),
            requires_grad=True
        )
        self.bias = torch.nn.Parameter(
            torch.zeros(1, n_drugs),
            requires_grad=True
        )

        self.vmap = None

    def forward(self, x, y=None):
        """
            Parameters
            ----------
            x: (batch_size, n_drugs) or (batch_size, )
                Doses matrix
            y: (batch_size, )
        """
        if self.non_linearity == 'logsigm':
            if y is None:
                c0 = self.bias.sigmoid()
                return (torch.log1p(x) * self.beta + self.bias).sigmoid() - c0
            else:
                bias = self.bias[0][y]
                beta = self.beta[0][y]
                c0 = bias.sigmoid()
                return (torch.log1p(x) * beta + bias).sigmoid() - c0
        elif self.non_linearity == 'sigm':
            if y is None:
                c0 = self.bias.sigmoid()
                return (x * self.beta + self.bias).sigmoid() - c0
            else:
                bias = self.bias[0][y]
                beta = self.beta[0][y]
                c0 = bias.sigmoid()
                return (x * beta + bias).sigmoid() - c0
        else:
            return x

    def one_drug(self, x, i):
        if self.non_linearity == 'logsigm':
            c0 = self.bias[0][i].sigmoid()
            return (torch.log1p(x) * self.beta[0][i] + self.bias[0][i]).sigmoid() - c0
        elif self.non_linearity == 'sigm':
            c0 = self.bias[0][i].sigmoid()
            return (x * self.beta[0][i] + self.bias[0][i]).sigmoid() - c0
        else:
            return x


class DrugNetwork(nn.Module):
    def __init__(self, n_drugs,
                 n_latent,
                 doser_type='logsigm',
                 n_hidden=None,
                 n_layers=None,
                 dropout_rate: float = 0.1):
        super().__init__()
        self.n_latent = n_latent
        self.drug_embedding = nn.Embedding(n_drugs, n_latent)
        self.doser_type = doser_type
        if self.doser_type == 'mlp':
            self.dosers = nn.ModuleList()
            for _ in range(n_drugs):
                self.dosers.append(
                    FCLayers(
                        n_in=1,
                        n_out=1,
                        n_hidden=n_hidden,
                        n_layers=n_layers,
                        use_batch_norm=False,
                        use_layer_norm=False,
                        dropout_rate=dropout_rate
                    )
                )
        else:
            self.dosers = GeneralizedSigmoid(n_drugs, non_linearity=self.doser_type)

    def forward(self, drugs, doses=None):
        """
            drugs: (batch_size, n_drugs) if combinatorial else (batch_size, )
                OneHot multiplied by doses if combinatorial is True
        """
        if self.doser_type == 'mlp':
            doses = []
            for d in range(drugs.size(1)):
                this_drug = drugs[:, d].view(-1, 1)
                doses.append(self.dosers[d](this_drug).sigmoid() * this_drug.gt(0))
            return torch.cat(doses, 1) @ self.drug_embedding.weight
        else:
            if doses is not None:
                drugs = drugs.long().view(-1, )
                doses = doses.float().view(-1, )
                scaled_dosages = self.dosers(doses, drugs)
                drug_embeddings = self.drug_embedding(drugs)
                return torch.einsum('b,be->be', [scaled_dosages, drug_embeddings])
            else:
                return self.dosers(drugs) @ self.drug_embedding.weight



import logging
import os
from typing import Optional, Sequence, Union, List

import numpy as np
import pandas as pd
import torch
from pytorch_lightning.callbacks import EarlyStopping
from scvi.data import AnnDataManager
from scvi.dataloaders import DataSplitter
from torch.nn import functional as F
from scvi.data.fields import (
    LayerField,
    CategoricalObsField,
    NumericalJointObsField,
    ObsmField,
)

from anndata import AnnData
from scvi.model.base import BaseModelClass
from scvi.train import TrainRunner
from scvi.train._callbacks import SaveBestState
from scvi.utils import setup_anndata_dsp
from tqdm import tqdm

from ._module import CPAModule
from ._utils import REGISTRY_KEYS
from ._task import CPATrainingPlan
from ._data import AnnDataSplitter

logger = logging.getLogger(__name__)
logger.propagate = False


class CPA(BaseModelClass):
    """CPA model

        Parameters
        ----------
        adata : Anndata
            Registered Annotation Dataset

        n_latent: int
            Number of latent dimensions used for drug and Autoencoder

        loss_ae: str
            Either `gauss` or `NB`. Autoencoder loss function.

        doser_type: str
            Type of doser network. Either `sigm`, `logsigm` or `mlp`.

        split_key : str, optional
            Key used to split the data between train test and validation.
            This must correspond to a observation key for the adata, composed of values
            'train', 'test', and 'ood'. By default None.

        **hyper_params:
            CPA's hyper-parameters.


        Examples
        --------
        >>> import cpa
        >>> import scanpy as sc
        >>> adata = sc.read('dataset.h5ad')
        >>> adata = cpa.CPA.setup_anndata(adata,
                                          drug_key='condition',
                                          dose_key='dose_val',
                                          categorical_covariate_keys=['cell_type'],
                                          control_key='control'
                                          )
        >>> hyperparams = {'autoencoder_depth': 3, 'autoencoder_width': 256}
        >>> model = cpa.CPA(adata,
                            n_latent=256,
                            loss_ae='gauss',
                            doser_type='logsigm',
                            split_key='split',
                            )
    """
    covars_encoder: dict = None
    drug_encoder: dict = None

    def __init__(
            self,
            adata: AnnData,
            n_latent: int = 128,
            loss_ae: str = 'gauss',
            doser_type: str = 'logsigm',
            split_key: str = None,
            train_split: str = 'train',
            valid_split: str = 'test',
            test_split: str = 'ood',
            **hyper_params,
    ):
        super().__init__(adata)
        self.drug_encoder = CPA.drug_encoder
        self.covars_encoder = CPA.covars_encoder

        self.n_genes = adata.n_vars
        self.n_drugs = len(self.drug_encoder)
        self.split_key = split_key

        self.drugs = list(self.drug_encoder.keys())
        self.covars = {
            covar: list(self.covars_encoder[covar].keys()) for covar in self.covars_encoder.keys()
        }

        self.module = CPAModule(
            n_genes=self.n_genes,
            n_drugs=self.n_drugs,
            covars_encoder=self.covars_encoder,
            n_latent=n_latent,
            loss_ae=loss_ae,
            doser_type=doser_type,
            **hyper_params,
        ).float()

        train_indices, valid_indices, test_indices = None, None, None
        if split_key is not None:
            train_indices = np.where(adata.obs.loc[:, split_key] == train_split)[0]
            valid_indices = np.where(adata.obs.loc[:, split_key] == valid_split)[0]
            test_indices = np.where(adata.obs.loc[:, split_key] == test_split)[0]

        self.train_indices = train_indices
        self.valid_indices = valid_indices
        self.test_indices = test_indices

        self._model_summary_string = f"Compositional Perturbation Autoencoder"

        self.init_params_ = self._get_init_params(locals())

        self.epoch_history = None

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
            cls,
            adata: AnnData,
            drug_key: str,
            dose_key: str,
            control_key: str,
            categorical_covariate_keys: Optional[List[str]] = None,
            continuous_covariate_keys: Optional[List[str]] = None,
            **kwargs,
    ):
        """
        Annotation Data setup function

        Parameters
        ----------
        adata

        drug_key

        dose_key

        control_key

        categorical_covariate_keys

        continuous_covariate_keys

        """
        REGISTRY_KEYS.DRUG_KEY = drug_key
        REGISTRY_KEYS.COVARS_KEYS = categorical_covariate_keys
        REGISTRY_KEYS.DOSE_KEY = dose_key
        REGISTRY_KEYS.CONTROL_KEY = control_key

        drugs = adata.obs[drug_key]
        dosages = adata.obs[dose_key].astype(str)

        # get unique drugs
        drugs_names_unique = set()
        for d in np.unique(drugs):
            [drugs_names_unique.add(i) for i in d.split("+")]
        drugs_names_unique = sorted(list(np.array(list(drugs_names_unique))))

        drugs_obsm = np.zeros((adata.n_obs, len(drugs_names_unique)))
        for i in tqdm(range(adata.n_obs)):
            cell_drugs = np.isin(drugs_names_unique, drugs[i].split('+'))
            cell_doses = np.array(dosages[i].split("+")).astype(np.float32)
            drugs_obsm[i, cell_drugs] = cell_doses

        adata.obsm['drugs_doses'] = np.array(drugs_obsm)

        drug_encoder = {drug: i for i, drug in
                        enumerate(drugs_names_unique)}

        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = \
            [
                LayerField(registry_key=REGISTRY_KEYS.X_KEY, layer=None, is_count_data=False),
                NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
                ObsmField('drugs_doses', 'drugs_doses', is_count_data=False, correct_data_format=True)
            ] + [CategoricalObsField(registry_key=covar, obs_key=covar) for covar in categorical_covariate_keys]

        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

        covars_encoder = {}
        for covar in categorical_covariate_keys:
            covars_encoder[covar] = {c: i for i, c in enumerate(
                adata_manager.registry['field_registries'][covar]['state_registry']['categorical_mapping'])}

        CPA.covars_encoder = covars_encoder
        CPA.drug_encoder = drug_encoder

    def train(
            self,
            max_epochs: Optional[int] = None,
            use_gpu: Optional[Union[str, int, bool]] = None,
            train_size: float = 0.9,
            validation_size: Optional[float] = None,
            batch_size: int = 128,
            early_stopping: bool = False,
            plan_kwargs: Optional[dict] = None,
            hyperopt: bool = False,
            save_path: Optional[str] = None,
            **trainer_kwargs,
    ):
        """
        Trains CPA on the given dataset

        Parameters
        ----------
        max_epochs: int
            Maximum number of epochs for training
        use_gpu: bool
            Whether to use GPU if available
        train_size: float
            Fraction of training data in the case of randomly splitting dataset to train/valdiation
                if `split_key` is not set in model's constructor
        validation_size: float
            Fraction of validation data in the case of randomly splitting dataset to train/valdiation
                if `split_key` is not set in model's constructor
        batch_size: int
            Size of mini-batches for training
        early_stopping: bool
            If `True`, EarlyStopping will be used during training on validation dataset
        plan_kwargs: dict
            `CPATrainingPlan` parameters
        save_path: str
            Path to save the model after the end of training
        """
        if max_epochs is None:
            n_cells = self.adata.n_obs
            max_epochs = np.min([round((20000 / n_cells) * 400), 400])
        plan_kwargs = plan_kwargs if isinstance(plan_kwargs, dict) else dict()

        manual_splitting = (
                (self.valid_indices is not None)
                and (self.train_indices is not None)
                and (self.test_indices is not None)
        )
        if manual_splitting:
            data_splitter = AnnDataSplitter(
                self.adata_manager,
                train_indices=self.train_indices,
                valid_indices=self.valid_indices,
                test_indices=self.test_indices,
                batch_size=batch_size,
                use_gpu=use_gpu,
            )
        else:
            data_splitter = DataSplitter(
                self.adata_manager,
                train_size=train_size,
                validation_size=validation_size,
                batch_size=batch_size,
                use_gpu=use_gpu,
            )

        self.training_plan = CPATrainingPlan(self.module, self.covars_encoder, **plan_kwargs)
        trainer_kwargs["early_stopping"] = False
        trainer_kwargs.update({'weights_summary': 'top'})
        trainer_kwargs['check_val_every_n_epoch'] = trainer_kwargs.get('check_val_every_n_epoch', 20)

        es_callback = EarlyStopping(monitor='cpa_metric',
                                    patience=trainer_kwargs['early_stopping_patience'],
                                    check_on_train_epoch_end=False,
                                    verbose=False,
                                    mode='max',
                                    )

        if 'callbacks' in trainer_kwargs.keys() and isinstance(trainer_kwargs.get('callbacks'), list):
            trainer_kwargs['callbacks'] += [es_callback]
        else:
            trainer_kwargs['callbacks'] = [es_callback]

        if save_path is None:
            save_path = './'

        checkpoint = SaveBestState(monitor='cpa_metric', mode='max', period=20, verbose=False)
        trainer_kwargs['callbacks'].append(checkpoint)

        runner = TrainRunner(
            self,
            training_plan=self.training_plan,
            data_splitter=data_splitter,
            max_epochs=max_epochs,
            use_gpu=use_gpu,
            early_stopping_monitor="cpa_metric",
            early_stopping_mode='max',
            checkpoint_callback=True,
            **trainer_kwargs,
        )
        runner()

        self.epoch_history = pd.DataFrame().from_dict(self.training_plan.epoch_history)
        if save_path is not None:
            self.save(save_path, overwrite=True)

    @torch.no_grad()
    def get_latent_representation(
            self,
            adata: Optional[AnnData] = None,
            indices: Optional[Sequence[int]] = None,
            batch_size: Optional[int] = 32,
    ) -> np.ndarray:
        """Returns the basal latent variable

        Parameters
        ----------
        adata : Optional[AnnData], optional
            [description], by default None
        indices : Optional[Sequence[int]], optional
            Optional indices, by default None
        batch_size : Optional[int], optional
            Batch size to use, by default None
        """

        if self.is_trained_ is False:
            raise RuntimeError("Please train the model first.")

        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )

        latent_basal = []
        latent = []
        for tensors in scdl:
            inference_inputs = self.module._get_inference_input(tensors)
            outputs = self.module.inference(**inference_inputs)
            latent_basal += [outputs["latent_basal"].cpu().numpy()]
            latent += [outputs['latent'].cpu().numpy()]

        latent_basal_adata = AnnData(X=np.concatenate(latent_basal, axis=0), obs=adata.obs.copy())
        latent_basal_adata.obs_names = adata.obs_names

        latent_adata = AnnData(X=np.concatenate(latent, axis=0), obs=adata.obs.copy())
        latent_adata.obs_names = adata.obs_names

        return latent_basal_adata, latent_adata

    @torch.no_grad()
    def predict(
            self,
            adata: Optional[AnnData] = None,
            indices: Optional[Sequence[int]] = None,
            batch_size: Optional[int] = 32,
    ):
        """Counterfactual-friendly gene expression prediction
        # TODO: See if another signature makes more sense for better usability

        To produce counterfactuals,

        Returns
        -------
        Tuple
            Gene expression means and standard variations
        """
        assert self.module.loss_ae in ["gauss", 'mse']
        self.module.eval()

        # adata = self.adata if adata is None else adata
        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )
        mus = []
        stds = []
        for tensors in scdl:
            _mus, _stds = self.module.get_expression(tensors)
            mus.append(_mus.detach().cpu().numpy())
            stds.append(_stds.detach().cpu().numpy())

        pred_adata_mean = AnnData(X=np.concatenate(mus, axis=0), obs=adata.obs.copy())
        pred_adata_var = AnnData(X=np.concatenate(stds, axis=0), obs=adata.obs.copy())

        pred_adata_mean.obs_names = adata.obs_names
        pred_adata_var.obs_names = adata.obs_names

        pred_adata_mean.var_names = adata.var_names
        pred_adata_var.var_names = adata.var_names

        return pred_adata_mean, pred_adata_var

    @torch.no_grad()
    def get_drug_embeddings(self, doses=1.0, drug: Optional[str] = None):
        """Computes all drug drug

        Parameters
        ----------
        doses : float, or torch.Tensor
            Drug dose, by default 1.0
        drug: str, optional
            Drug name if single drug embedding is desired

        """
        self.module.eval()
        if isinstance(doses, float):
            if drug is None:
                treatments = doses * torch.eye(self.n_drugs, device=self.device)
            else:
                treatments = doses * F.one_hot(torch.LongTensor([self.drug_encoder[drug]]).to(self.device),
                                               self.n_drugs)
        elif isinstance(doses, np.ndarray):
            treatments = torch.tensor(doses).to(self.device).float()
        else:
            treatments = doses

        embeds = self.module.drug_network(treatments).detach().cpu().numpy()

        return embeds

    @torch.no_grad()
    def get_covar_embeddings(self, covariate: str, covariate_value: str = None):
        """Computes Covariate drug

        Parameters
        ----------
        covariate : str
            covariate to be computed
        covariate_value: str, Optional
            Covariate specific value for embedding computation

        """
        if covariate_value is None:
            covar_ids = torch.arange(len(self.covars_encoder[covariate]), device=self.device)
        else:
            covar_ids = torch.LongTensor([self.covars_encoder[covariate][covariate_value]]).to(self.device)
        embeddings = self.module.covars_embedding[covariate](covar_ids).detach().cpu().numpy()

        return embeddings

    def save(self, dir_path: str, overwrite: bool = False, save_anndata: bool = False, **anndata_write_kwargs):
        os.makedirs(dir_path, exist_ok=True)

        if isinstance(self.training_plan.epoch_history, dict):
            self.epoch_history = pd.DataFrame().from_dict(self.training_plan.epoch_history)
            self.epoch_history.to_csv(os.path.join(dir_path, 'history.csv'), index=False)

        return super().save(dir_path=dir_path, overwrite=overwrite, save_anndata=save_anndata, **anndata_write_kwargs)

    @classmethod
    def load(cls, dir_path: str, adata: Optional[AnnData] = None, use_gpu: Optional[Union[str, int, bool]] = None):
        model = super().load(dir_path, adata, use_gpu)
        CPA.covars_encoder = model.covars_encoder
        CPA.drug_encoder = model.drug_encoder

        try:
            model.epoch_history = pd.read_csv(os.path.join(dir_path, 'history.csv'))
        except:
            print('WARNING: The history was not found.')

        return model
